from base64 import b64decode

import regex

YTSM = regex.compile(r"ysmm = '([^']+)")


def decode_adfly(coded_key):
    r, j = "", ""
    for n, l in enumerate(coded_key):
        if n & 1:
            r += l
        else:
            j = l + j

    encoded_uri = list(r + j)
    numbers = ((i, n) for i, n in enumerate(encoded_uri) if str.isdigit(n))
    for first, second in zip(numbers, numbers):
        xor = int(first[1]) ^ int(second[1])
        if xor < 10:
            encoded_uri[first[0]] = str(xor)

    return b64decode(("".join(encoded_uri)).encode("utf-8"))[16:-16].decode(
        "utf-8", errors="ignore"
    )


def bypass_adfly(session, adfly_url):

    response_code = 302

    while response_code != 200:
        adfly_content = session.get(
            session.get(adfly_url, follow_redirects=False).headers.get("location"),
            follow_redirects=False,
        )
        response_code = adfly_content.status_code
    return decode_adfly(YTSM.search(adfly_content.text).group(1))



import functools
import time

import pypresence

ANIMDL_APPLICATION_ID = 897850955373105222

try:
    presence_client = pypresence.Presence(ANIMDL_APPLICATION_ID)
except (pypresence.DiscordNotFound, ConnectionError):
    raise RuntimeError("Could not connect to Discord.")

presence_client.connect()

presence_client.update(
    state="Queuing up the streams",
    large_image="mascot",
    large_text="https://github.com/justfoolingaround/animdl",
)

KITSU_ENDPOINT = "https://kitsu.io/api/"


@functools.lru_cache()
def get_anime(session, anime_name):
    return session.get(
        KITSU_ENDPOINT + "edge/anime", params={"filter[text]": anime_name}
    ).json()["data"]


@functools.lru_cache()
def get_episodes(session, anime_id, offset):

    params = {
        "filter[media_id]": anime_id,
        "sort": "number",
        "filter[mediaType]": "Anime",
        "page[limit]": 10,
    }

    if offset is not None and offset > 1:
        params["page[offset]"] = offset

    return session.get(KITSU_ENDPOINT + "edge/episodes", params=params).json()["data"]


def set_streaming_episode(session, anime_name, episode):

    state = anime_name

    if episode:
        state += " - Episode {}".format(episode)

    content_list = get_anime(session, anime_name)

    if not content_list:
        return presence_client.update(
            state=state,
            large_image="mascot",
            large_text="https://github.com/justfoolingaround/animdl",
            start=int(time.time()),
        )

    anime = content_list[0]

    anime_attributes = anime["attributes"]

    image = anime_attributes["posterImage"]["original"]
    count = anime_attributes["episodeCount"]
    url = "https://kitsu.io/{}/{}".format(anime['type'], anime_attributes['slug'])
    url_button = [ { "label": "View Anime", "url": url } ]

    if count:
        state += "/{}".format(count)

    if count is None or (episode > count):
        return presence_client.update(
            state=state,
            large_image=image,
            large_text=anime_name,
            small_image="mascot",
            small_text="https://github.com/justfoolingaround/animdl",
            start=int(time.time()),
            buttons=url_button,
        )

    around = episode - (episode % 10)

    current = get_episodes(session, anime["id"], around)[episode % 10 - 1]

    episode_name = current.get("attributes", {}).get("canonicalTitle")
    episode_thumbnail = (current.get("attributes", {}).get("thumbnail", {}) or {}).get(
        "original", "mascot"
    )

    if episode_thumbnail == "mascot":
        image, episode_thumbnail = episode_thumbnail, image

    return presence_client.update(
        state=state,
        large_image=episode_thumbnail,
        large_text=episode_name if image != "mascot" else anime_name,
        small_image=image,
        small_text=anime_name if image != "mascot" else "https://github.com/justfoolingaround/animdl",
        details=episode_name,
        start=int(time.time()),
        buttons=url_button,
    )



import json

import click

from ...__version__ import __core__
from ...codebase import providers
from ...config import CHECK_FOR_UPDATES, DEFAULT_PROVIDER
from .. import helpers
from ..http_client import client


@click.command(
    name="grab", help="Stream the stream links to the stdout stream for external usage."
)
@helpers.decorators.content_fetch_options(
    include_quality_options=False,
    include_special_options=False,
)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_grab(query, index, **kwargs):

    console = helpers.stream_handlers.get_console()
    console.print(
        "The content is outputted to [green]stdout[/] while these messages are outputted to [red]stderr[/]."
    )

    anime, provider = helpers.process_query(
        client, query, console, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        return

    for stream_url_caller, episode in providers.get_appropriate(
        client, anime.get("anime_url"), check=kwargs.get("range")
    ):
        stream_url = list(helpers.ensure_extraction(client, stream_url_caller))
        click.echo(json.dumps({"episode": episode, "streams": stream_url}))



import subprocess
import sys
from collections import defaultdict

import click
import yarl
from rich.prompt import Prompt
from rich.text import Text

from ...config import FZF_EXECUTABLE, FZF_OPTS, FZF_STATE
from .intelliq import filter_quality
from .stream_handlers import context_raiser


def default_prompt(
    console,
    components,
    *,
    processor=None,
    component_name="result",
    fallback=None,
    error_message=None,
    stdout_processor=None,
    escape_output=False,
):

    with context_raiser(
        console, f"[b]Waiting for you to select a {component_name!r}.[/]"
    ):
        if processor is None:
            components = list(components)
        else:
            components = list(processor(component) for component in components)

        if not components:
            return fallback

        if len(components) == 1:
            return components[0]

        processed_components = list(
            stdout_processor(component) if stdout_processor is not None else component
            for component in components
        )

        enumerated_components = tuple(enumerate(processed_components, 1))

        for n, cout in reversed(enumerated_components):
            if not isinstance(cout, str):
                raise TypeError(
                    "The stdout_processor must return a string and not {!r}.".format(
                        type(cout)
                    )
                )
            console.print(Text(f"{n}.", style="b blue"), cout)

        choices = list(map(str, range(1, len(enumerated_components) + 1)))

        choice = Prompt.ask(
            Text(
                f"Select the {component_name} (automatically selects the top {component_name}) ",
                style="dim",
            )
            + Text(f'[{"/".join(choices)}]', style="b blue"),
            default=1,
            console=console,
            choices=choices,
            show_choices=False,
        )

    return components[int(choice) - 1]


def fzf_prompt(
    console,
    components,
    *,
    processor=None,
    component_name="result",
    fallback=None,
    error_message=None,
    stdout_processor=None,
    escape_output=False,
):
    if processor is None:
        components = list(components)
    else:
        components = list(processor(component) for component in components)

    if len(components) == 1:
        return components[0]

    stdout_mapout = {}

    for component_output, component in zip(
        (stdout_processor(component) for component in components)
        if stdout_processor
        else components,
        components,
    ):

        if not isinstance(component_output, str):
            raise TypeError(
                "The stdout_processor must return a string and not {!r}.".format(
                    type(component_output)
                )
            )

        if escape_output:
            component_output = repr(component_output)[1:-1]

        if component_output in stdout_mapout:
            component_output += " (apparent duplicate {})".format(component_name)

        stdout_mapout[component_output] = component

    if not stdout_mapout:
        if error_message is not None:
            console.print(error_message, style="bold red")
        return fallback

    fzf_args = [
        FZF_EXECUTABLE,
        "--header={}".format(
            f"Select the {component_name} (automatically selects the top {component_name})"
        ),
    ] + FZF_OPTS

    process = subprocess.Popen(
        fzf_args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    selection, _ = process.communicate(b"\n".join(map(str.encode, stdout_mapout)))

    if process.returncode in [1, 130]:
        return components[0]

    return stdout_mapout[selection.decode()[:-1]]


def get_prompt_manager(*, fallback=default_prompt):

    if not sys.stdout.isatty():
        return fallback

    if FZF_STATE:
        return fzf_prompt

    return fallback


def quality_prompt(console, streams: list[dict], *, force_selection_string=None):

    if len(streams) == 1:
        return streams[0]

    if force_selection_string is not None:
        return quality_prompt(console, filter_quality(streams, force_selection_string))

    def stream_stdout_processor(content: dict):
        parsed_url = yarl.URL(content["stream_url"])

        title = content.get("title")

        if title is None:
            title = f"{parsed_url.name} // {parsed_url.host!r}"

        if content.get("subtitle"):
            title += " (with subtitles)"

        quality = content.get("quality")

        if quality is not None:
            title += f" [{quality}p]"
        else:
            extension = parsed_url.suffix.lstrip(".")

            if extension in ("m3u8", "mpd", "m3u"):
                title += f" [Adaptive quality: {extension}]"
            else:
                title += f" [Unknown quality: {extension}]"

        return title

    value = get_prompt_manager()(
        console,
        streams,
        stdout_processor=stream_stdout_processor,
        component_name="stream",
    )

    return value



from base64 import b64decode
from functools import partial

import lxml.html as htmlparser
import regex

from ....config import ANIMEXIN
from ...helpers import construct_site_based_regex

REGEX = construct_site_based_regex(
    ANIMEXIN,
    extra_regex=r"/(?:anime/(?P<normal_slug>[^&?#/]+)|(?P<slug>.+?)-episode-\d+-.+?)",
)


TITLES_REGEX = regex.compile(r'<h2 itemprop="partOfSeries">(.+?)</h2>')
IFRAME_SRC = regex.compile(r'src="(.+?)"')

FURTHER_EXTRACTORS = {
    "gdriveplayer": "//gdriveplayer.to",
    "dailymotion": "https://www.dailymotion.com/",
}


def get_further_extractor_name(url):
    for extractor_name, prefix in FURTHER_EXTRACTORS.items():
        if isinstance(prefix, regex.Pattern) and prefix.search(url):
            return extractor_name
        if url.startswith(prefix):
            return extractor_name
    return None


def get_episode_metadata(episode_element):

    generic_episode_number = 0
    generic_episode_name = None

    e_n = episode_element.cssselect("div.epl-num")
    e_t = episode_element.cssselect("div.epl-title")

    if e_n:
        number = e_n[0].text_content()
        if number.isdigit():
            generic_episode_number = int(number)

    if e_t:
        generic_episode_name = (
            e_t[0].text_content()
            if not generic_episode_number
            else "Episode {} - {}".format(generic_episode_number, e_t[0].text_content())
        )

    return generic_episode_number, generic_episode_name


def extract(session, url, **opts):
    for options in htmlparser.fromstring(session.get(url).text).cssselect(
        "select.mirror > option[data-index][value]"
    ):
        embed_uri_match = IFRAME_SRC.search(b64decode(options.get("value")).decode())
        if not embed_uri_match:
            continue

        embed_uri = embed_uri_match.group(1)
        further_ext = get_further_extractor_name(embed_uri)

        if not further_ext or further_ext in ["gdriveplayer"]:
            continue

        yield {"stream_url": embed_uri, "further_extraction": (further_ext, {}), **opts}


def fetcher(session, url, check, match):

    if match.group("slug"):
        url = ANIMEXIN + "anime/{}".format(match.group("slug"))

    for episode in htmlparser.fromstring(session.get(url).text).cssselect(
        ".eplister li[data-index] > a"
    )[::-1]:
        number, name = get_episode_metadata(episode)

        if check(number):
            yield partial(
                lambda s, u, **kwargs: list(extract(s, u, **kwargs)),
                session,
                episode.get("href"),
                title=name,
            ), number


def metadata_fetcher(session, url, match):

    if match.group("slug"):
        url = ANIMEXIN + "anime/{}".format(match.group("slug"))

    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



from ...codebase.providers import get_provider
from .prompts import get_prompt_manager
from .searcher import provider_searcher_mapping


def prompt_user(logger, anime_list_genexp, provider):

    manager = get_prompt_manager()

    return manager(
        logger,
        anime_list_genexp,
        processor=lambda component: (component, provider),
        component_name="search result",
        fallback=({}, None),
        error_message=f"Failed to find anything of that query on {provider!r}. Try searching on other providers.",
        stdout_processor=lambda component: f"{component[0]['name']} / {component[0]['anime_url']}",
    )


def process_query(session, query: str, console, provider: str, *, auto_index=1):

    match, module, provider_name = get_provider(query, raise_on_failure=False)

    if module:
        return {
            "anime_url": query,
            "name": module.metadata_fetcher(session, query, match).get(
                "titles", [None]
            )[0],
        }, provider_name

    provider_name, *custom_query = query.split(":", 1)

    if provider_name in provider_searcher_mapping:
        provider, query = provider_name, ":".join(custom_query)

    genexp = provider_searcher_mapping[provider](session, query)

    if auto_index is None:
        return prompt_user(console, genexp, provider)

    expanded = list(genexp)
    return expanded[(auto_index - 1) % len(expanded)], provider



import logging

import click


class ColouredFormatter(logging.Formatter):

    logging_colours = {
        "WARNING": "yellow",
        "INFO": "green",
        "DEBUG": "blue",
        "CRITICAL": "red",
        "ERROR": "red",
    }

    def format(self, record: logging.LogRecord) -> str:

        foreground = self.logging_colours.get(record.levelname, "white")

        record.levelname = click.style(record.levelname, fg=foreground)
        record.name = click.style(record.name, fg=foreground)

        return super().format(record)


class FileLoggingFormatter(logging.Formatter):

    logging_format = (
        "[%(filename)s:%(lineno)d - %(asctime)s - %(name)s: %(levelname)s] %(message)s"
    )

    def __init__(self):
        super().__init__(fmt=self.logging_format)


def setup_loggers(
    logging_format=f"[{click.style('%(filename)s:%(lineno)d', fg='magenta')} - %(asctime)s - %(name)s: %(levelname)s] %(message)s",
    file_logging_format=f"[%(filename)s:%(lineno)d - %(asctime)s - %(name)s: %(levelname)s] %(message)s",
):
    def wrapper(f):
        def __inner__(*args, log_level=100, log_file=None, **kwargs):

            basic_config_kwargs = {
                "format": logging_format,
                "level": log_level,
            }

            if log_file is not None:
                basic_config_kwargs.update(
                    filename=log_file,
                    filemode="a",
                )

            logging.basicConfig(**basic_config_kwargs)

            class SmartColouredLogger(logging.Logger):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                    self.propagate = False

                    if log_file is not None:
                        file_stream_handler = logging.FileHandler(
                            log_file, "a", encoding="utf-8"
                        )
                        file_stream_handler.setFormatter(
                            logging.Formatter(fmt=file_logging_format)
                        )
                        self.addHandler(file_stream_handler)
                    else:
                        stream_handler = logging.StreamHandler()
                        stream_handler.setFormatter(
                            ColouredFormatter(fmt=logging_format)
                        )
                        self.addHandler(stream_handler)

            logging.setLoggerClass(SmartColouredLogger)

            return f(*args, **kwargs, log_file=log_file, log_level=log_level)

        return __inner__

    return wrapper



"""
Useless since there is no open-source way to bypass ouo.io (currently).

This is quite slow as it tries to get all the streams at once.
"""

from base64 import b64decode
from functools import partial

import regex

from ....config import ANIMEKAIZOKU
from ...helpers import construct_site_based_regex, group_episodes, parse_from_content

REGEX = construct_site_based_regex(ANIMEKAIZOKU, extra_regex=r"/([^?&/]+)")


ON_NEW_TAB = regex.compile(r'openInNewTab\("(.+?)"\)\'><p>Download (.+?)</p>')

ANIMEKAIZOKU_DDL = regex.compile(r"DDL\((.+?), (.+?), '(.+?)', (.+?)\)")

DDL_DIVID = regex.compile(r"glist-(\d+)")
DDL_POSTID = regex.compile(r'"postId":"(\d+)"')

TITLES_REGEX = regex.compile(r'<h1 class="post-title entry-title">(.+?)</h1>')


def graceful_ajax(session, data):
    return session.post(
        ANIMEKAIZOKU + "wp-admin/admin-ajax.php",
        headers={"x-requested-with": "XMLHttpRequest", "referer": ANIMEKAIZOKU},
        data=data,
    )


def walk(session, post_id, div_id, tab_id, num, folder):

    loaded_page = graceful_ajax(
        session,
        {
            "action": "DDL",
            "post_id": post_id,
            "div_id": div_id,
            "tab_id": tab_id,
            "num": num,
            "folder": folder,
        },
    ).text

    for match in ANIMEKAIZOKU_DDL.finditer(loaded_page):
        yield from walk(session, post_id, *match.groups())

    for match in ON_NEW_TAB.finditer(loaded_page):
        yield match


def fetcher(session, url, check, match):

    response = session.get(url).text

    div_id = DDL_DIVID.search(response).group(1)
    post_id = DDL_POSTID.search(response).group(1)

    for episode, streams in sorted(
        group_episodes(
            (
                parse_from_content(
                    regex_match,
                    name_processor=lambda m: m.group(2),
                    stream_url_processor=lambda m: (
                        b64decode(m.group(1)).decode("utf-8")
                    ),
                )
                for regex_match in walk(session, post_id, div_id, 2, "", True)
            )
        ).items()
    ):
        if check(episode):
            yield partial(list, streams), episode


def metadata_fetcher(session, url: "str", match):
    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



import base64
from textwrap import wrap
from urllib.parse import quote, unquote

NORMAL_TABLE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"


def cipher_keyed(key, plaintext):
    cipher = ""
    xcrypto = 0

    mapping = {_: _ for _ in range(256)}

    for c in range(256):
        xcrypto = (xcrypto + mapping[c] + ord(key[c % len(key)])) % 256
        mapping[c], mapping[xcrypto] = mapping[xcrypto], mapping[c]

    i = j = 0
    for f in range(0, len(plaintext)):
        j = (j + 1) % 256  # NOTE: Previously, this was 'j = (j + f) % 256'
        i = (i + mapping[j]) % 256
        mapping[i], mapping[j] = mapping[j], mapping[i]
        cipher += chr(ord(plaintext[f]) ^ mapping[(mapping[i] + mapping[j]) % 256])

    return cipher


def get_salted_code(plaintext):
    part_1 = encrypt((unquote(plaintext) + "000000")[:6])[0:4][::-1]
    return part_1 + encrypt(cipher_keyed(part_1, unquote(plaintext)))[0:4].replace(
        "=", ""
    )


def encrypt(data, *, table=NORMAL_TABLE):
    return "".join(
        table[int(segment.ljust(6, "0"), 2) if len(segment) < 6 else int(segment, 2)]
        for segment in wrap(
            "".join(bin(ord(c)).lstrip("0b").rjust(8, "0") for c in data), 6
        )
    )


def decrypt_url(encrypted_url, secret):
    return unquote(cipher_keyed(secret, decrypt(encrypted_url)))


def encrypt_url(url, key):
    return encrypt(cipher_keyed(key, quote(url)))


def decrypt(data, *, table=NORMAL_TABLE):
    return "".join(
        map(
            chr,
            base64.b64decode(
                (data + "=" * (len(data) % 4))
                .translate(str.maketrans(table, NORMAL_TABLE))
                .encode()
            ),
        )
    )



from sys import platform
from tempfile import NamedTemporaryFile

from .base_player import BasePlayer


class MPVDefaultPlayer(BasePlayer):

    optimisation_args = [
        "--force-window=immediate",
    ]
    headers_joiner = "\r\n"

    if platform == "win32":
        path_joiner = ";"
    else:
        path_joiner = ":"

    opts_spec = {
        "headers": "--http-header-fields",
        "title": "--title",
        "subtitles": "--sub-files",
        "media-title": "--force-media-title",
    }

    def play(
        self,
        stream_url,
        subtitles=None,
        headers=None,
        title=None,
        opts=None,
        chapters=None,
        **kwargs,
    ):

        args = (self.executable, stream_url)

        if opts is not None:
            args += tuple(opts)

        if headers is not None:
            args += (
                f"{self.opts_spec['headers']}={self.headers_joiner.join(f'{key}: {value}' for key, value in headers.items())}",
            )

        if title is not None:
            args += (
                f"{self.opts_spec['title']}={title}",
                f"{self.opts_spec['media-title']}={title}",
            )

        if subtitles is not None:
            args += (
                f"{self.opts_spec['subtitles']}={self.path_joiner.join(subtitles)}",
            )

        if chapters:

            # NOTE: This could be achieved with a PIPE.
            # This is not done in this case because you
            # only PIPE one argument at a time, and this
            # is expected to be limiting in the future.
            # PIPE-ing whole media is more suitable.

            with NamedTemporaryFile("w", delete=False) as chapters_file:
                chapters_file.write(";FFMETADATA1")

                for chapter in chapters:
                    chapters_file.write(
                        f"""
[CHAPTER]
TIMEBASE=1/1000
START={int(chapter["start"] * 1000)}
END={int(chapter["end"] * 1000)}
TITLE={chapter["chapter"]}"""
                    )

                chapters_file.write("\n")
            args += (f"--chapters-file={chapters_file.name}",)

        args += tuple(self.optimisation_args)

        self.spawn(args)


class CelluloidPlayer(MPVDefaultPlayer):
    optimisation_args = [
        "--new-window",
    ]

    def __new__(cls, *args, **kwargs):

        for key, value in cls.opts_spec.items():
            cls.opts_spec[key] = f"--mpv-{value.lstrip('-')}"

        return super().__new__(cls)


IINAPlayer = CelluloidPlayer



import base64
import logging

import click
import yarl
from rich.text import Text

from ...__version__ import __core__
from ...codebase import providers
from ...config import (
    CHECK_FOR_UPDATES,
    DEFAULT_PLAYER,
    DEFAULT_PROVIDER,
    DISCORD_PRESENCE,
    FORCE_STREAMING_QUALITY_SELECTION,
    QUALITY,
    USE_ANISKIP,
)
from .. import exit_codes, helpers, http_client

if DISCORD_PRESENCE:
    try:
        from ..helpers.rpc import set_streaming_episode
    except ImportError:
        raise ImportError(
            "Discord RPC was set to be enabled but `pypresence` is not installed, install it with `pip install pypresence`."
        )
    except RuntimeError:
        DISCORD_PRESENCE = False


@click.command(name="stream", help="Stream your favorite anime by query.")
@helpers.decorators.content_fetch_options(
    default_quality_string=QUALITY,
)
@helpers.decorators.player_options(default_player=DEFAULT_PLAYER)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    http_client.client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_stream(
    query, special, quality, player_opts, player, index, log_level, **kwargs
):
    """
    Streamer call for animdl streaming session.
    """
    r = kwargs.get("range")
    console = helpers.stream_handlers.get_console()

    logger = logging.getLogger("streamer")

    if player_opts is not None:
        player_opts = tuple(click.parser.split_arg_string(player_opts))
    else:
        player_opts = ()

    try:
        streamer = helpers.player.handle_player(DEFAULT_PLAYER, player_opts, player)
    except Exception:
        console.print_exception()
        raise SystemExit(exit_codes.STREAMER_CONFIGURATION_REQUIRED)

    anime, provider = helpers.process_query(
        http_client.client, query, console, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        console.print(Text("Could not find an anime of that name :/."))
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    with helpers.stream_handlers.context_raiser(
        console,
        Text(
            f"Scraping juicy streams from {provider!r}@{anime['anime_url']}",
            style="bold magenta",
        ),
        name="scraping",
    ):

        match, provider_module, _ = providers.get_provider(
            providers.append_protocol(anime.get("anime_url"))
        )

        streams = list(
            provider_module.fetcher(
                http_client.client, anime.get("anime_url"), r, match
            )
        )

        if special:
            streams = list(helpers.special.special_parser(streams, special))

        content_title = anime["name"]

        total = len(streams)

        if total < 1:
            console.print(Text("Could not find any streams on the site :/."))
            console.print(
                "This could mean that, either those episodes are unavailable or that the scraper has broke.",
                style="dim",
            )
            return exit(exit_codes.NO_CONTENT_FOUND)

        with helpers.stream_handlers.context_raiser(
            console, f"Now streaming {content_title!r}", name="streaming"
        ):

            for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

                if episode_number == 0:
                    episode_text = "Special Episode"
                else:
                    episode_text = f"Episode {episode_number}"

                if content_title:
                    media_title = f"{content_title}: {episode_text}"
                else:
                    media_title = episode_text

                playing = True

                while playing:

                    stream_urls = list(
                        helpers.ensure_extraction(
                            http_client.client, stream_urls_caller
                        )
                    )

                    if not stream_urls:

                        console.print(
                            Text(
                                f"Could not find any streams for {episode_text!r} :/."
                            ),
                            Text(
                                "Loading in DEBUG [--log-level=0] mode allows extraction retrial.",
                                style="dim",
                            ),
                        )

                        if log_level <= logging.DEBUG:
                            playing = click.confirm("Retry stream url extraction?")

                        continue

                    titles = set(_.get("title") for _ in stream_urls)

                    if len(titles) > 1 and FORCE_STREAMING_QUALITY_SELECTION:
                        console.print(
                            Text(
                                f"Multiple titles found for the same streams, multiple seasons may be available. Please adjust the quality string as required: {', '.join(map(repr, titles))}",
                                style="dim",
                            ),
                        )

                    selection = helpers.prompts.quality_prompt(
                        console,
                        stream_urls,
                        force_selection_string=quality
                        if FORCE_STREAMING_QUALITY_SELECTION
                        else None,
                    )

                    headers = selection.get("headers", {})

                    logger.debug(f"Obtained streams: {stream_urls!r}")

                    if "title" in selection:
                        media_title += f" ({selection['title']})"

                    if USE_ANISKIP:
                        chapters = helpers.aniskip.get_timestamps(
                            http_client.client, content_title, episode_number
                        )
                    else:
                        chapters = []

                    with helpers.stream_handlers.context_raiser(
                        console,
                        f"Currently playing: {episode_text!r}. [dim]{total-count:d} episodes in queue.[/]",
                        name="playing",
                    ):
                        if not headers:
                            shareable_url = yarl.URL(
                                "https://plyr.link/p/player.html"
                            ).with_fragment(
                                base64.b64encode(
                                    selection["stream_url"].encode()
                                ).decode()
                            )
                            console.print(
                                Text(
                                    f"This stream may be share-able [embeddable and playable as a single URL] (please note that subtitles, chapters and titles won't be embedded)."
                                ),
                                style="dim",
                            )
                            console.print(
                                f"[dim]Share-ables: [link={shareable_url}]embed[/], [link={selection['stream_url']}]direct url[/][/]"
                            )

                        with streamer:
                            streamer.play(
                                selection["stream_url"],
                                title=media_title,
                                headers=headers,
                                subtitles=selection.get("subtitle", []),
                                chapters=chapters,
                            )

                            if DISCORD_PRESENCE:
                                set_streaming_episode(
                                    http_client.client, content_title, episode_number
                                )

                    error_indication = streamer.indicate_error()

                    if bool(error_indication):
                        logger.warning(
                            f"Streamer indicated an error {error_indication!r}."
                        )

                        if log_level <= logging.INFO:
                            if click.confirm("Retry streaming this episode?"):
                                continue

                    if log_level <= logging.DEBUG:
                        if click.confirm("Replay episode?"):
                            continue

                    playing = False



import logging
from pathlib import Path

import click
from rich.text import Text

from ...__version__ import __core__
from ...codebase import providers, sanitize_filename
from ...config import (
    AUTO_RETRY,
    CHECK_FOR_UPDATES,
    DEFAULT_PROVIDER,
    DOWNLOAD_DIRECTORY,
    QUALITY,
)
from .. import exit_codes, helpers, http_client


@click.command(name="download", help="Download your favorite anime by query.")
@helpers.decorators.content_fetch_options(
    default_quality_string=QUALITY,
)
@helpers.decorators.download_options(
    default_download_dir=DOWNLOAD_DIRECTORY,
)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    http_client.client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_download(
    query, special, quality, download_dir, idm, index, log_level, **kwargs
):
    r = kwargs.get("range")

    console = helpers.stream_handlers.get_console()

    logger = logging.getLogger("downloader")

    anime, provider = helpers.process_query(
        http_client.client, query, console, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        console.print(Text("Could not find an anime of that name :/."))
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    logger.name = "{}/{}".format(provider, logger.name)

    match, provider_module, _ = providers.get_provider(
        providers.append_protocol(anime.get("anime_url"))
    )

    with helpers.stream_handlers.context_raiser(
        console,
        Text(
            f"Scraping juicy streams from {provider!r}@{anime['anime_url']}",
            style="bold magenta",
        ),
        name="scraping",
    ):
        streams = list(
            provider_module.fetcher(
                http_client.client, anime.get("anime_url"), r, match
            )
        )

        if special:
            streams = list(helpers.special.special_parser(streams, special))

        download_directory = Path(download_dir).resolve(strict=True)
        content_name = sanitize_filename(anime["name"])

        content_dir = download_directory / content_name
        content_dir.mkdir(exist_ok=True)

        console.print(
            "The project will download to:",
            Text(content_dir.resolve().as_posix(), style="bold"),
        )
        total = len(streams)

        if total < 1:
            console.print(Text("Could not find any streams on the site :/."))
            console.print(
                "This could mean that, either those episodes are unavailable or that the scraper has broke.",
                style="dim",
            )
            raise SystemExit(exit_codes.NO_CONTENT_FOUND)

        with helpers.stream_handlers.context_raiser(
            console, f"Now downloading {content_name!r}", name="downloading"
        ):
            for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

                content_title = "E{:02d}".format(int(episode_number))
                stream_urls = helpers.ensure_extraction(
                    http_client.client, stream_urls_caller
                )

                if not stream_urls:
                    console.print(
                        Text(
                            f"Could not find any streams for {content_title!r} :/.",
                        )
                    )
                    continue

                console.print(
                    f"Currently downloading: {content_title!r}. [dim]{total-count:d} episodes in queue.[/]",
                )

                success, reason = helpers.download(
                    http_client.client,
                    logger,
                    content_dir,
                    content_title,
                    stream_urls,
                    quality,
                    idm=idm,
                    retry_timeout=AUTO_RETRY,
                    log_level=log_level,
                )

                if not success:
                    console.print(
                        Text(
                            f"Could not download {content_title!r} because of {reason!r} :/, a good option would be to retry using better providers :/.",
                        )
                    )



import functools

import regex

DAILYMOTION_VIDEO = regex.compile(r"/embed/video/([^&?/]+)")
HLS_STREAM_REGEX = regex.compile(r'NAME="(\d+?)",PROGRESSIVE-URI="(.+?)"')


def get_mp4s(session, m3u8_playlist):
    for match in HLS_STREAM_REGEX.finditer(
        session.get(
            m3u8_playlist, headers={"Referer": "https://www.dailymotion.com/"}
        ).text
    ):
        yield {
            "quality": int(match.group(1)),
            "stream_url": match.group(2),
        }


def extract(session, url, **opts):
    match = DAILYMOTION_VIDEO.search(url)

    if not match:
        return []

    metadata_uri = "https://www.dailymotion.com/player/metadata/video/{}".format(
        match.group(1)
    )

    metadata = session.get(metadata_uri).json()

    subtitles = functools.reduce(
        list.__add__,
        (list(data.values()) for _, data in metadata.get("subtitles", {}).get("data")),
        [],
    )

    def genexp():

        for _, streams in metadata.get("qualities", {}).items():
            for stream in streams:
                yield from (
                    list(get_mp4s(session, stream.get("url")))
                    or [{"stream_url": stream.get("url")}]
                )

    return [
        {
            **stream,
            **({"subtitles": subtitles} if subtitles else {}),
        }
        for stream in genexp()
    ]



import functools

import regex

from ....config import ANIMEPAHE
from ...helpers import construct_site_based_regex
from .inner import get_animepahe_url

REGEX = construct_site_based_regex(ANIMEPAHE, extra_regex=r"/(anime|play)/([^?&/]+)")

ID_RE = regex.compile(r'let id = "(.+?)"')
KWIK_RE = regex.compile(r"Plyr\|(.+?)'")

TITLES_REGEX = regex.compile(r"<h1>(.+?)</h1>")


def get_streams_from_embed_url(session, embed_uri):
    embed_page = session.get(embed_uri, headers={"referer": ANIMEPAHE}).text
    return "{}://{}-{}.{}.{}.{}/{}/{}/{}/{}.{}".format(
        *KWIK_RE.search(embed_page).group(1).split("|")[::-1]
    )


def iter_stream_url_from_stream_session(session, release_id, stream_session):

    stream_url_data = session.get(
        ANIMEPAHE + "api",
        params={"m": "links", "id": release_id, "session": stream_session, "p": "kwik"},
    )

    for qualities in stream_url_data.json().get("data", []):
        for quality, data in qualities.items():
            yield {
                "quality": quality,
                "stream_url": get_animepahe_url(session, data["kwik_pahewin"]),
            }


def iter_episode_streams(session, release_id, per_page, episode_number):

    current_page = episode_number // per_page + 1

    episode = fetch_session(session, release_id, page=current_page)["data"][
        episode_number % per_page - 1
    ]

    yield from iter_stream_url_from_stream_session(
        session, release_id, episode["session"]
    )


def bypass_ddos_guard(session):
    js_bypass_uri = regex.search(
        r"'(.*?)'", session.get("https://check.ddos-guard.net/check.js").text
    ).group(1)
    session.cookies.update(session.get(ANIMEPAHE + js_bypass_uri).cookies)


@functools.lru_cache()
def fetch_session(session, release_id, *, page=None):
    return session.get(
        ANIMEPAHE + "api",
        params={"m": "release", "id": release_id, "sort": "episode_asc", "page": page},
    ).json()


def fetcher(session, url, check, match):

    if match.group(1) == "play":
        url = ANIMEPAHE + f"anime/{match.group(2)}"

    anime_page = session.get(url)

    release_id = ID_RE.search(anime_page.text).group(1)

    initial_session = fetch_session(session, release_id, page=1)
    per_page = initial_session["per_page"]
    total = initial_session["total"]

    for episode_number in range(1, total + 1):
        if check(episode_number):
            yield functools.partial(
                (
                    lambda episode_number: iter_episode_streams(
                        session, release_id, per_page, episode_number
                    )
                ),
                episode_number,
            ), episode_number


def metadata_fetcher(session, url, match):
    if match.group(1) == "play":
        url = ANIMEPAHE + f"anime/{match.group(2)}"

    return {
        "titles": TITLES_REGEX.findall(session.get(url).text),
    }



from .fuzzysearch import search

ENDPOINT = "https://api.aniskip.com/v1"

MAL_XHR_SEARCH = "https://myanimelist.net/search/prefix.json"


type_keywords = {
    "op": "Opening",
    "ed": "Ending",
}


def iter_general_timestamps(aniskip_data):

    op_end = None
    ed_start = None

    for item in aniskip_data:

        skip_type = item["skip_type"]

        if skip_type == "op":
            op_end = item["interval"]["end_time"]

        if skip_type == "ed":
            ed_start = item["interval"]["start_time"]

        yield {
            "chapter": type_keywords.get(skip_type, skip_type),
            "start": item["interval"]["start_time"],
            "end": item["interval"]["end_time"],
        }

    if op_end is not None:
        yield {
            "chapter": "Episode",
            "start": op_end,
            "end": ed_start or op_end,
        }

    # NOTE: This is a continuation fix, mpv does not seem to like unmarked chapters between two chapters.


def get_timestamps(session, anime_name, anime_episode):

    data = (
        session.get(
            MAL_XHR_SEARCH,
            params={"type": "anime", "keyword": anime_name},
        )
        .json()
        .get("categories", [{}])[0]
        .get("items", [])
    )

    if not data:
        return

    top_result = (
        tuple(search(anime_name, data, processor=lambda _: _["name"])) or data
    )[0]

    ani_skip_response = session.get(
        f"{ENDPOINT}/skip-times/{top_result['id']}/{anime_episode}",
        params={
            "types[]": ("op", "ed"),
        },
    )

    if ani_skip_response.status_code < 400:
        json_data = ani_skip_response.json()
        if json_data["found"]:
            return list(iter_general_timestamps(json_data["results"]))



"""
IntelliQ, a highly intelligent and robust quality string parser.
"""


import logging
from typing import Callable, Generator, Iterable, List, Optional, Tuple, TypeVar, Union

import regex

PORTION_PARSER = regex.compile(r'(.+?)=(r?)("|\')((?:\\\3|.)*?)\3')
SEGMENT_PARSER = regex.compile("(best|worst|\d+)?(.*)")

pair_value = TypeVar("pair_value")
stream_type = TypeVar("stream_type")


def NO_PROCESS(stream: stream_type):
    return stream


def get_pair(
    target: "pair_value", pairs: Tuple[pair_value, pair_value]
) -> Tuple[Union[Tuple[pair_value, pair_value], Tuple[None, None], bool]]:

    for f, l in pairs:
        if target in (f, l):
            return (f, l), target == f

    return (None, None), False


def parse_parenthesized_portions(
    segment: "str",
) -> Tuple[str, Optional[Union[str, regex.Pattern]]]:
    match = PORTION_PARSER.search(segment)

    if match is None:
        return (segment, None)

    if match.group(2):
        return (match.group(1), regex.compile(match.group(4)))

    return (match.group(1), match.group(4))


def portion_check(
    portions: Iterable[Tuple[str, Optional[Union[str, regex.Pattern]]]]
) -> Callable[[stream_type], bool]:

    for key, portion in portions:
        if portion is None:
            yield lambda stream, k=key: bool(stream.get(k))
            continue

        if isinstance(portion, regex.Pattern):
            yield lambda stream, k=key, p=portion: bool(
                p.search(str(stream.get(k, "")))
            )
            continue

        yield lambda stream, k=key, p=portion: stream.get(k, "") == p


def parenthesized_portions(
    string: "str",
    escape: "str" = "\\",
    quoters: List[str] = ["'", '"'],
    parenthesis: List[Tuple[str, str]] = [("[", "]"), ("(", ")"), ("{", "}")],
) -> Generator[Tuple[str, Optional[Union[str, regex.Pattern]]], None, None]:

    initiator, endpoint = min(
        parenthesis, key=lambda x: (string.find(x[0]) + 1) or float("inf")
    )

    escaping = False
    current_context = ""

    multiquote_context = dict.fromkeys(quoters, False)

    pos = string.find(initiator)

    if pos == -1:
        return

    for current_pos, content in enumerate(string[pos + 1 :], pos + 1):

        if not escaping:
            if content in quoters:
                multiquote_context[content] = not multiquote_context[content]

            if content == endpoint and not any(multiquote_context.values()):

                yield parse_parenthesized_portions(current_context.strip())
                yield from parenthesized_portions(
                    string[current_pos:],
                    escape=escape,
                    quoters=quoters,
                    parenthesis=parenthesis,
                )
                return

        current_context += content
        escaping = content in escape


def split_portion(
    string: str,
    splitters: List[str] = ["/"],
    escape: "str" = "\\",
    quoters: List[str] = ["'", '"'],
    parenthesis: List[Tuple[str, str]] = [("[", "]"), ("(", ")"), ("{", "}")],
) -> Generator[str, None, None]:

    multiquote_context = dict.fromkeys(quoters, False)
    parenthesis_context = dict.fromkeys(parenthesis, False)

    escaping = False
    current_context = ""
    yield_this_loop = False

    for content in string:

        pair, is_initiator = get_pair(content, parenthesis)

        if not escaping:
            if content in quoters:
                multiquote_context[content] = not multiquote_context[content]

            if pair in parenthesis_context:
                if not parenthesis_context[pair]:
                    if is_initiator:
                        parenthesis_context[pair] = True
                else:
                    if not is_initiator:
                        parenthesis_context[pair] = False

            if (
                content in splitters
                and not any(multiquote_context.values())
                and not any(parenthesis_context.values())
            ):
                yield current_context.strip()
                yield_this_loop = True

        if yield_this_loop:
            current_context = ""
        else:
            current_context += content

        escaping = content in escape
        yield_this_loop = False

    yield current_context.strip()


def get_int(key: Optional[Union[int, str]]) -> int:

    if key is None:
        return 0

    if isinstance(key, int):
        return key

    if isinstance(key, str) and key.isdigit():
        return int(key)

    digits = regex.search(key, r"[0-9]+")

    if digits:
        return int(digits.group(0))

    return 0


def parse_quality_only(quality: str) -> Callable[[stream_type], bool]:
    if quality == "best":
        return lambda streams: [
            max(streams, key=lambda stream: get_int(stream.get("quality", 0)))
        ]

    if quality == "worst":
        return lambda streams: [
            min(streams, key=lambda stream: get_int(stream.get("quality", 0)))
        ]

    if quality and not quality.isdigit():
        return NO_PROCESS

    return lambda streams: sorted(
        list(
            stream
            for stream in streams
            if get_int(stream.get("quality", 0)) <= float(quality or "inf")
        ),
        key=lambda stream: get_int(stream.get("quality", 0)),
        reverse=True,
    )


def finalise_check(
    quality_check: Callable[[stream_type], bool],
    parsed_parenthesized_portions: "Tuple[str, Optional[Union[str, regex.Pattern]]]",
) -> Callable[[stream_type], List[stream_type]]:
    def internal(streams):
        streams = list(
            stream
            for stream in streams
            if all(_(stream) for _ in portion_check(parsed_parenthesized_portions))
        )

        if not streams:
            return []

        return quality_check(streams)

    return internal


def parse_quality_string(
    quality_string: str,
) -> Generator[
    Tuple[str, Callable[[stream_type], Optional[List[stream_type]]]], None, None
]:

    quality_string = quality_string.lower()

    for segment in split_portion(quality_string):
        match = SEGMENT_PARSER.search(segment)
        yield segment, finalise_check(
            parse_quality_only(match.group(1)),
            list(parenthesized_portions(match.group(2))),
        )


def filter_quality(
    streams: Iterable[stream_type], quality_string: str
) -> List[stream_type]:

    logger = logging.getLogger("utils/intelliq")

    logger.debug("Parsing {!r} in {!r}".format(quality_string, streams))

    for segment, check in parse_quality_string(quality_string):
        filtered = check(streams)

        if filtered:
            logger.debug("{} streams fulfill {!r}.".format(len(filtered), segment))
            return filtered

        logger.debug("No streams fulfill {!r}.".format(segment))

    logger.debug("Quality checks have failed miserably. Returning everything back.")
    return streams



import functools
from functools import partial

import regex

from ....config import ANIMTIME
from ...helpers import construct_site_based_regex

CONTENT_RE = regex.compile(r't\[t\.(?P<name>.+?)=(?P<id>\d+)\]="\1"')
REGEX = construct_site_based_regex(ANIMTIME, extra_regex=r"/title/([^/?&]+)")

MAIN_JS_FILE = "main.0eab142b791320871345.js"


def get_content(content_id, js_content):
    for content in CONTENT_RE.finditer(js_content):
        if content.group("id") == content_id:
            return content.group("name")


@functools.lru_cache()
def get_js_content(session):
    return session.get(ANIMTIME + MAIN_JS_FILE).text


def fetcher(session, url, check, match):

    content = get_js_content(session)[374391:]

    anime = get_content(match.group(1), content)

    if anime is None:
        return

    episode_count = int(
        regex.search(rf"Ld\[Kd\.{regex.escape(anime)}\]=(\d+)", content).group(1)
    )

    constructor, end = regex.search(
        regex.escape(f"Fd[Kd.{anime}]=function(t){{return")
        + '"(.+?)"'
        + regex.escape("+t+")
        + r'"(.+?)"\}',
        content,
    ).groups()

    for episode in range(1, episode_count + 1):
        if check(episode):
            yield partial(
                lambda x: [{"stream_url": x, "headers": {"referer": ANIMTIME}}],
                constructor + "{:03d}".format(episode) + end,
            ), episode


def metadata_fetcher(session, url, match):
    content = get_js_content(session)[374391:]

    anime_name = get_content(match.group(1), content)

    if anime_name is not None:
        anime_name = " ".join(regex.split(r"(?=[A-Z])", anime_name)).strip()

    return {"titles": [anime_name]}



try:
    import regex as regexlib
except ImportError:
    import re as regexlib

try:
    import orjson as jsonlib

    dumps_function = jsonlib.dumps

    def patched_dumps(*args, **kwargs):
        return dumps_function(*args, **kwargs).decode()

    jsonlib.dumps = patched_dumps

except ImportError:
    import json as jsonlib



import shutil
from warnings import warn

from ...config import PLAYERS
from .players import player_mapping


def iter_available_players():
    for player, player_info in PLAYERS.items():

        executable = player_info.get("executable")

        if executable is None or shutil.which(executable) is None:
            continue

        yield player, player_info


def get_player(default_player, user_selected_player=None):

    players = tuple(iter_available_players())

    if not players:
        raise RuntimeError(
            "No players found on the system. "
            "Make sure the player is installed and in PATH or in the current directory or has a configured executable."
        )

    if user_selected_player is None:
        user_selected_player = default_player

    for player, player_info in players:
        if player == user_selected_player:
            return player, player_info

    fallback_player, _ = players[0]

    warn(
        f"Could not find {user_selected_player!r} in the system. "
        f"Falling back to the first available player {fallback_player!r}."
    )

    return fallback_player, _


def handle_player(default_player: str, player_opts: tuple, user_selected_player=None):
    if default_player is None:
        raise RuntimeError("No default player set in the configuration.")

    player, player_info = get_player(default_player, user_selected_player)

    cls = player_mapping[player]

    return cls(
        executable=player_info["executable"],
        args=tuple(player_info.get("opts", ())) + player_opts,
    )



"""
A one cli for all the anime.
"""

import os
import random
import sys

import click
from rich.align import Align
from rich.style import Style
from rich.text import Text
from rich.traceback import install

from .core import __version__
from .core.cli.commands import download, grab, schedule, search, stream, update
from .core.cli.helpers import stream_handlers

install(show_locals=True, suppress=[click])

commands = {
    "download": download.animdl_download,
    "grab": grab.animdl_grab,
    "schedule": schedule.animdl_schedule,
    "search": search.animdl_search,
    "update": update.animdl_update,
    "stream": stream.animdl_stream,
}


@click.group(commands=commands, invoke_without_command=True)
@click.version_option(__version__.__core__, "--version")
@click.option("-x", "--disable-update", is_flag=True, help="Disable update check.")
@click.pass_context
def __animdl_cli__(ctx: click.Context, disable_update):
    from .core.cli.helpers import constants

    sys.stderr = sys.__stderr__
    console = stream_handlers.get_console()

    if ctx.invoked_subcommand is None:
        console.print(
            "The project is operational. Use --help flag for help & be sure read the documentation!"
        )

    author, repository_name = constants.SOURCE_REPOSITORY

    console.print(
        Align(
            Text(
                f"{author}/{repository_name} v{__version__.__core__}",
                style="magenta",
            ),
            align="center",
        )
    )

    with stream_handlers.context_raiser(
        console,
        Text(
            f"{random.choice(('Hey', 'Hello', 'Welcome'))}, {os.getenv('USERNAME', 'buddy')}.",
            style="bold white",
        ),
    ):
        for greeting in stream_handlers.iter_greetings():
            console.print(greeting, style="white")

        if not disable_update:
            from .core.cli.http_client import client
            from .core.codebase.helpers.optopt import regexlib

            branch, version_file = constants.VERSION_FILE_PATH

            upstream_version = regexlib.search(
                r'__core__ = "(.*?)"',
                client.get(
                    f"https://raw.githubusercontent.com/{author}/{repository_name}/{branch}/{version_file}"
                ).text,
            ).group(1)

            tuplisied_upstream, tuplised_current_version = tuple(
                map(int, upstream_version.split("."))
            ), tuple(map(int, __version__.__core__.split(".")))

            if tuplisied_upstream > tuplised_current_version:
                with stream_handlers.context_raiser(
                    console, "[cyan]An update is available![/]"
                ):
                    console.print(
                        f"↑ {upstream_version} ↓ {__version__.__core__}, to update, use: animdl update",
                        style=Style(color="magenta"),
                    )
                    console.print(
                        f"Staying up-to-date resolves bugs and security vulnerabilities!",
                        style="dim i",
                    )


if __name__ == "__main__":
    __animdl_cli__()



import functools
import time

API_URL = "https://api.kamyroll.tech"


class Kamyroll:

    API_URL = "https://api.kamyroll.tech/"

    def __init__(self, session):

        self.session = session
        self.raw_token = {}

    def get_token(self):
        expires = self.raw_token.get("expires_in", 0)

        if expires < time.time():

            response = self.session.get(
                self.API_URL + "auth/v1/token",
                params={
                    "device_id": "web",
                    "device_type": "python.animdl",
                    "access_token": "HMbQeThWmZq4t7w",
                },
            )

            self.raw_token = response.json()
        else:
            return f"{self.raw_token['token_type']} {self.raw_token['access_token']}"

        return self.get_token()

    def fetch_seasons(self, media_id, *, channel_id="crunchyroll", locale="en-US"):
        return self.session.get(
            self.API_URL + "content/v1/seasons",
            headers={
                "Authorization": self.get_token(),
            },
            params={"channel_id": channel_id, "id": media_id, "locale": locale},
        ).json()

    def fetch_streams(
        self,
        media_id,
        *,
        channel_id="crunchyroll",
        locale="en-US",
        media_type="adaptive_hls",
    ):
        return self.session.get(
            self.API_URL + "videos/v1/streams",
            params={
                "channel_id": channel_id,
                "locale": locale,
                "id": media_id,
                "type": media_type,
            },
            headers={
                "Authorization": self.get_token(),
            },
        ).json()

    def iter_search_results(self, query, *, channel_id="crunchyroll", locale="en-US"):
        results = self.session.get(
            self.API_URL + "content/v1/search",
            params={
                "channel_id": channel_id,
                "limit": 250,
                "query": query,
            },
            headers={
                "Authorization": self.get_token(),
            },
        ).json()["items"]

        for medias in results:
            for search_results in medias["items"]:
                yield {
                    "title": search_results["title"],
                    "media_id": search_results["id"],
                    "type": medias["type"],
                }

    def fetch_media(self, media_id, *, channel_id="crunchyroll", locale="en-US"):
        return self.session.get(
            self.API_URL + "content/v1/media",
            params={
                "channel_id": channel_id,
                "id": media_id,
                "locale": locale,
            },
            headers={
                "Authorization": self.get_token(),
            },
        ).json()


@functools.lru_cache()
def get_api(session):
    return Kamyroll(session)



import functools

import lxml.html as htmlparser

from ....config import HAHO
from ...helpers import construct_site_based_regex, optopt

REGEX = construct_site_based_regex(HAHO, extra_regex=r"/anime/([^?&/]+)")

TITLES_REGEX = optopt.regexlib.compile(r'<h1 class="mb-3">(.+?)</h1>')
SORUCES_REGEX = optopt.regexlib.compile(
    r'<source src="(.+?)" title="([0-9]+?)p" .+?>',
    flags=optopt.regexlib.I,
)


def iter_stream_urls(session, episode_page):
    episode_page_content = session.get(episode_page)

    iframes = htmlparser.fromstring(episode_page_content.text).cssselect("iframe")

    if not iframes:
        return

    embed_source = session.get(
        iframes[0].get("src"), headers={"referer": episode_page}
    ).text

    for source in SORUCES_REGEX.finditer(embed_source):
        yield {
            "stream_url": source.group(1),
            "quality": int(source.group(2)),
        }


def fetcher(session, url, check, match):
    url = match.group(0)

    episode_list_page = session.get(url)
    count = int(
        htmlparser.fromstring(episode_list_page.text)
        .cssselect("span.badge")[0]
        .text_content()
    )

    for episode in range(1, count + 1):
        if check(episode):
            yield functools.partial(
                (lambda url: list(iter_stream_urls(session, url))), f"{url}/{episode}"
            ), episode


def metadata_fetcher(session, url, match):

    url = match.group(0)

    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



import functools

import lxml.html as htmlparser
import regex

from ....config import MARIN
from ...helpers import construct_site_based_regex, optopt

REGEX = construct_site_based_regex(MARIN, extra_regex=r"/anime/([^?&/]+)")

inertia_headers = {
    "x-inertia-version": "67023a9481a03976f4d5ae9b16e8f752",
    "x-inertia": "true",
}


class HashableDict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def iter_episode_streams(session, url):

    episodes_data = session.get(
        url,
        headers=inertia_headers,
    ).json()

    props = episodes_data.get("props", {})

    for episode in props.get("video", {}).get("data", {}).get("mirror", []):
        code = episode.get("code", {})

        yield {
            "stream_url": code.get("file"),
            "quality": code.get("width"),
            "title": " // ".join(
                _["text"]
                for _ in props.get("episode", {}).get("data", {}).get("title")[:2]
            ),
        }


@functools.lru_cache()
def fetch_anime_data(session, url):
    return HashableDict(
        session.get(
            url,
            headers=inertia_headers,
        ).json()
    )


def fetcher(session, url, check, match):
    url = match.group(0)

    anime_data = fetch_anime_data(session, url)

    count = int(anime_data.get("props", {}).get("anime", {}).get("last_episode", 0))

    for episode in range(1, count + 1):
        if check(episode):
            yield functools.partial(
                lambda url: list(iter_episode_streams(session, url)),
                f"{url}/{episode}",
            ), episode


def metadata_fetcher(session, url, match):

    url = match.group(0)

    anime_data = fetch_anime_data(session, url)

    return {
        "titles": [anime_data.get("props", {}).get("anime", {}).get("title")],
    }



"""
All the search algorithms for all the providers available in AnimDL.
"""

import json
from urllib.parse import unquote

import lxml.html as htmlparser

from ...codebase.helpers import uwu
from ...codebase.providers.kamyroll.api import get_api
from ...config import (
    ALLANIME,
    ANIMEKAIZOKU,
    ANIMEOUT,
    ANIMEPAHE,
    ANIMIXPLAY,
    GOGOANIME,
    HAHO,
    HENTAISTREAM,
    KAWAIIFU,
    MARIN,
    NINEANIME,
    TWIST,
    YUGEN,
    ZORO,
)
from .fuzzysearch import search


def search_9anime(session, query):
    nineanime_results = session.get(NINEANIME + "filter", params={"keyword": query})

    parsed = htmlparser.fromstring(nineanime_results.text)
    for results in parsed.cssselect("a.name"):
        yield {
            "anime_url": NINEANIME.rstrip("/") + results.get("href"),
            "name": results.text_content(),
        }


def search_animekaizoku(session, query):
    animekaizoku_results = htmlparser.fromstring(
        session.get(ANIMEKAIZOKU, params={"s": query}).text
    )
    for results in animekaizoku_results.cssselect(".post-title"):
        yield {
            "anime_url": ANIMEKAIZOKU + results.cssselect("a")[0].get("href"),
            "name": results.text_content(),
        }


def search_allanime(session, query):

    gql_response = session.get(
        ALLANIME + "allanimeapi",
        params={
            "variables": json.dumps(
                {
                    "search": {
                        "allowAdult": True,
                        "allowUnknown": True,
                        "query": query,
                    },
                    "limit": 40,
                }
            ),
            "extensions": json.dumps(
                {
                    "persistedQuery": {
                        "version": 1,
                        "sha256Hash": "9c7a8bc1e095a34f2972699e8105f7aaf9082c6e1ccd56eab99c2f1a971152c6",
                    }
                }
            ),
        },
    ).json()

    for result in gql_response.get("data", {}).get("shows", {}).get("edges", []):
        if any(a for _, a in result.get("availableEpisodes", {}).items()):
            yield {
                "anime_url": ALLANIME + f"anime/{result['_id']}",
                "name": result["name"],
            }


def search_animepahe(session, query):

    animepahe_results = session.get(
        ANIMEPAHE + "api", params={"q": query, "m": "search"}
    )
    content = animepahe_results.json()

    for results in content.get("data", []):
        yield {
            "anime_url": ANIMEPAHE + "anime/" + results.get("session"),
            "name": results.get("title"),
        }


def search_animeout(session, query):
    animeout_results = session.cf_request("GET", ANIMEOUT, params={"s": query})
    content = htmlparser.fromstring(animeout_results.text)

    for result in content.xpath('//h3[@class="post-title entry-title"]/a'):
        yield {"anime_url": result.get("href"), "name": result.text_content()}


def search_animixplay(session, query):
    animixplay_library = (
        _ for _ in session.get(ANIMIXPLAY + "assets/s/all.json").json() if _["e"] == "1"
    )

    for result in search(query, animixplay_library, processor=lambda r: r.get("title")):
        yield {
            "anime_url": ANIMIXPLAY + "v1/" + result.get("id"),
            "name": result.get("title"),
        }


def search_gogoanime(session, query):
    parsed = htmlparser.fromstring(
        session.get(GOGOANIME + "/search.html", params={"keyword": query}).text
    )

    for results in parsed.xpath('//p[@class="name"]/a'):
        yield {
            "anime_url": GOGOANIME.strip("/") + results.get("href"),
            "name": results.get("title"),
        }


def search_kawaiifu(session, query):
    for results in htmlparser.fromstring(
        session.get(KAWAIIFU + "search-movie", params={"keyword": query}).text
    ).cssselect(".info > h4 > a:last-child"):
        yield {"anime_url": results.get("href"), "name": results.text_content().strip()}


def search_twist(session, query):
    content = session.get(
        "https://api.twist.moe/api/anime",
        headers={"x-access-token": "0df14814b9e590a1f26d3071a4ed7974"},
    )
    animes = content.json()

    for anime in search(
        query, animes, processor=lambda r: r.get("title") or r.get("alt_title")
    ):
        yield {
            "anime_url": TWIST + "a/" + anime.get("slug", {}).get("slug"),
            "name": anime.get("title", ""),
        }


def search_crunchyroll(session, query, *, scheme="http"):
    # ajax/?req=RpcApiSearch_GetSearchCandidates

    kamyroll = get_api(session)
    for anime in kamyroll.iter_search_results(query):
        yield {
            "anime_url": f"{scheme}://www.crunchyroll.com/{anime['type']}/{anime['media_id']}",
            "name": anime["title"],
        }


def search_marin(session, query, *, domain=MARIN):

    session.get(domain, headers={"range": "bytes=0-0"})

    for result in (
        session.post(
            domain + "anime",
            json={
                "search": query,
            },
            headers={
                "x-inertia": "true",
                "x-xsrf-token": unquote(session.cookies.get("XSRF-TOKEN")),
            },
        )
        .json()
        .get("props", {})
        .get("anime_list", {})
        .get("data", [])
    ):
        yield {
            "name": result.get("title"),
            "anime_url": domain + "anime/" + result.get("slug"),
        }


def search_yugen(session, query):
    for result in search(
        query,
        htmlparser.fromstring(
            session.get(YUGEN + "discover/", params={"dq": query}).text
        ).cssselect("a.anime-meta[href][title]"),
        processor=lambda r: r.get("title"),
    ):
        yield {
            "name": result.get("title"),
            "anime_url": YUGEN + result.get("href")[1:],
        }


def search_zoro(session, query):
    for result in htmlparser.fromstring(
        session.get(ZORO + "/search", params={"keyword": query}).text
    ).cssselect("a.item-qtip[title][data-id]"):
        yield {
            "name": result.get("title"),
            "anime_url": ZORO + result.get("href")[1:-11],
        }


def search_h_ntai_stream(session, query):
    for result in htmlparser.fromstring(
        session.get(HENTAISTREAM + "search/", params={"s": query}).text
    ).cssselect("article > .bsx > a"):
        yield {
            "name": result.get("title") or result.get("oldtitle"),
            "anime_url": HENTAISTREAM + result.get("href")[1:],
        }


def search_haho(session, query):
    tenshi_page = htmlparser.fromstring(
        session.get(HAHO + "anime", params={"q": query}).text
    )

    for result in tenshi_page.cssselect(".list > li > a"):
        yield {"name": result.get("title"), "anime_url": result.get("href")}


provider_searcher_mapping = {
    "9anime": search_9anime,
    "animekaizoku": search_animekaizoku,
    "allanime": search_allanime,
    "animepahe": search_animepahe,
    "animeout": search_animeout,
    "animixplay": search_animixplay,
    "crunchyroll": search_crunchyroll,
    "kamyroll": lambda *args, **kwargs: search_crunchyroll(
        *args, **kwargs, scheme="kamyroll"
    ),
    "kawaiifu": search_kawaiifu,
    "gogoanime": search_gogoanime,
    "haho": search_haho,
    "hentaistream": search_h_ntai_stream,
    "marin": search_marin,
    "twist": search_twist,
    "yugen": search_yugen,
    "zoro": search_zoro,
}



import regex

KWIK_PARAMS_RE = regex.compile(r'\("(\w+)",\d+,"(\w+)",(\d+),(\d+),\d+\)')
KWIK_D_URL = regex.compile(r'action="(.+?)"')
KWIK_D_TOKEN = regex.compile(r'value="(.+?)"')

KWIK_REDIRECTION_RE = regex.compile(r'<a href="(.+?)" .+?>Redirect me</a>')

CHARACTER_MAP = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+/"


def get_string(content, s1, s2):

    slice_2 = CHARACTER_MAP[0:s2]

    acc = 0
    for n, i in enumerate(content[::-1]):
        acc += int(i if i.isdigit() else 0) * s1**n

    k = ""
    while acc > 0:
        k = slice_2[int(acc % s2)] + k
        acc = (acc - (acc % s2)) / s2

    return k or "0"


def decrypt(full_string, key, v1, v2):
    v1, v2 = int(v1), int(v2)
    r = ""
    i = 0
    while i < len(full_string):
        s = ""
        while full_string[i] != key[v2]:
            s += full_string[i]
            i += 1
        j = 0
        while j < len(key):
            s = s.replace(key[j], str(j))
            j += 1
        r += chr(int(get_string(s, v2, 10)) - v1)
        i += 1
    return r


def get_animepahe_url(session, pahe_win_url):

    response = session.get(pahe_win_url)
    response.raise_for_status()

    url = KWIK_REDIRECTION_RE.search(response.text).group(1)

    download_page = session.get(url)

    full_key, key, v1, v2 = KWIK_PARAMS_RE.search(download_page.text).group(1, 2, 3, 4)

    decrypted = decrypt(
        full_key,
        key,
        v1,
        v2,
    )

    content = session.post(
        KWIK_D_URL.search(decrypted).group(1),
        follow_redirects=False,
        data={"_token": KWIK_D_TOKEN.search(decrypted).group(1)},
        headers={
            "Referer": "https://kwik.cx/",
        },
    )

    return content.headers["Location"]



import asyncio
import pathlib
import shlex
import shutil
import sys

import click
import regex

from ...__version__ import __core__
from .. import helpers

GIT_CONFIG_REPO_NAME = regex.compile(r'\[remote ".+?"\].+?url = (\S+)', flags=regex.S)


def is_repository(author, repository):

    if not pathlib.Path("./.git/").exists():
        return False

    repository_qualname = f"{author}/{repository}"

    with open("./.git/config") as git_config:
        match = GIT_CONFIG_REPO_NAME.search(git_config.read())

        if match is None:
            return False

    return match.group(1)[-len(repository_qualname) :] == repository_qualname


@click.command("update", help="Update the project or git pull if a repository.")
@helpers.decorators.setup_loggers()
def animdl_update(*args, **kwargs):

    console = helpers.stream_handlers.get_console()

    with helpers.stream_handlers.context_raiser(
        console,
        helpers.stream_handlers.Text(
            f"Updating {helpers.constants.MODULE_NAME!r}",
            style="bold white",
        ),
        name="update",
    ):

        executable = None

        async def __asyncio__():

            if is_repository(*helpers.constants.SOURCE_REPOSITORY):

                executable = shutil.which("git")
                args = [
                    executable,
                    "pull",
                ]

                console.print(
                    f"Pulling latest changes from the repository via git: {shlex.join(args)}"
                )

                if not executable:
                    return console.print("[red]Cannot find git.[/]")

                process = await asyncio.subprocess.create_subprocess_exec(
                    *args,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )

            else:

                executable = sys.executable

                args = [
                    executable,
                    "-m",
                    "pip",
                    "install",
                    "--upgrade",
                    helpers.constants.MODULE_NAME,
                    "--user",
                    "--no-warn-script-location",
                ]

                console.print(
                    f"Updating module {helpers.constants.MODULE_NAME!r} via PIP: {shlex.join(args)}"
                )

                process = await asyncio.subprocess.create_subprocess_exec(
                    *args,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )

            async def async_stream_reader(stream, style, name):

                line = await stream.readline()
                while line:
                    console.print(
                        helpers.stream_handlers.Text(name, style=style),
                        helpers.stream_handlers.Text(line.decode().strip()),
                    )
                    line = await stream.readline()

            await asyncio.gather(
                async_stream_reader(process.stdout, "green", f"[{executable} stdout]"),
                async_stream_reader(process.stderr, "red", f"[{executable} stderr]"),
            )

        loop = asyncio.new_event_loop()
        loop.run_until_complete(__asyncio__())



import datetime
import functools

from rich.console import Console
from rich.style import Style
from rich.text import Text


@functools.lru_cache()
def get_console():
    return Console(stderr=True, color_system="auto")


class ContextRaiser:
    def __init__(self, console, message=None, *, indent_string=" " * 2, name="unknown"):
        self.indent_string = indent_string
        self.console = console
        self.default_writer = console.print
        self.message = message

        self.ctx_name = name

    def __enter__(self):
        if self.message:
            self.console.print(self.message)

        def print(*args, **kwargs):
            self.default_writer(self.indent_string, *args, **kwargs)

        self.console.print = print

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type:
            if isinstance(exc_value, (KeyboardInterrupt, SystemExit)):
                self.console.print(
                    f"Exiting from the {self.ctx_name} context.",
                    style="bold red",
                )

        self.console.print = self.default_writer


context_raiser = ContextRaiser


def is_day(date, month, day):
    return date.month == month and date.day == day


custom_greetings = {
    lambda date: is_day(date, 1, 1): Text("Happy new year! 🎉", style="bold yellow"),
    lambda date: is_day(date, 12, 25): Text("Merry Christmas! 🎄", style="bold cyan"),
    lambda date: is_day(date, 4, 1): Text(
        "Happy April Fools' Day! 🎉", style="bold green"
    ),
    lambda date: is_day(date, 10, 31): Text(
        "Ooh, spooky! Happy Halloween. 🎃", style="yellow"
    ),
    lambda date: is_day(date, 3, 14): Text("Happy Valentine's Day ❤!", style="red"),
}


def iter_greetings():

    datetime_now = datetime.datetime.now()

    for condition, message in custom_greetings.items():
        if condition(datetime_now):
            yield message

    yield datetime_now.strftime("It is %I:%M %p on a [bold red]beautiful[/] %A!")



from collections import defaultdict
from functools import partial

import lxml.html as htmlparser
import regex
import yarl

REGEX = regex.compile(
    r"(?:https?://)?(?:\S+\.)?(?P<host>domdom\.stream|bestanime3\.xyz|kawaiifu\.com)/(?P<episode_page>anime/)?(?P<type>season/[^/]+|.+)/(?P<slug>[^?&#]+)"
)
TITLES_REGEX = regex.compile(r'<h2 class="title">(.+?)</h2>')


def get_int(content):
    d = regex.search(r"[0-9]+", content)
    if d:
        return int(d.group(0))


def extract_stream_urls(session, urls):
    for url in urls:
        html_element = htmlparser.fromstring(session.get(url).text)
        for source in html_element.cssselect("source"):
            yield {
                "quality": get_int(source.get("data-quality")),
                "stream_url": source.get("src"),
                "headers": {"referer": url},
            }


def get_from_url(session, url):
    episodes = defaultdict(list)
    html_element = htmlparser.fromstring(session.get(url).text)

    for servers in html_element.cssselect(".list-server"):
        for element in servers.cssselect(".list-ep a"):
            episodes[get_int(element.text_content()) or 0].append(element.get("href"))
    return episodes


def fetcher(session, url, check, match):

    url = yarl.URL(url).with_host("bestanime3.xyz").human_repr()

    for episode, episode_urls in sorted(
        get_from_url(session, url).items(), key=lambda x: x[0]
    ):
        if check(episode):
            yield partial(
                lambda episode_urls: list(extract_stream_urls(session, episode_urls)),
                episode_urls,
            ), episode


def metadata_fetcher(session, url, match):
    return {
        "titles": TITLES_REGEX.findall(
            session.get(yarl.URL(url).with_host("bestanime3.xyz").human_repr()).text
        )
    }



import base64
from functools import partial

from ....config import YUGEN
from ...helpers import construct_site_based_regex, optopt, superscrapers

REGEX = construct_site_based_regex(
    YUGEN, extra_regex=r"/(?:watch|anime)/(?P<anime_path>(?P<anime_id>.+?)/.+?)(?:/|$)"
)


EPISODE_REGEX = optopt.regexlib.compile(
    r'<div class="ap-.+?">Episodes</div><span class="description" .+?>(\d+)</span></div>'
)
EPISODE_TITLE_REGEX = optopt.regexlib.compile(r'<h1 class="p-10-t">\n(.+?)\n</h1>')


def iter_streams(session, anime_id, episode_id):

    data = (
        session.post(
            YUGEN + "api/embed/",
            data={
                "id": base64.b64encode(f"{anime_id}|{episode_id}".encode()).decode(),
                "ac": "0",
            },
            headers={
                "x-requested-with": "XMLHttpRequest",
            },
        )
        .json()
        .get("hls", [])
    )

    for source in data:
        url = superscrapers.yarl.URL(source)

        yield from superscrapers.iter_unpacked_from_packed_hls(session, url)


def fetcher(session, url, check, match):

    anime_page = YUGEN + "anime/" + match.group("anime_path")
    anime_page_html = session.get(anime_page).text
    subbed_episode_count = int(EPISODE_REGEX.search(anime_page_html).group(1))

    for episode in range(1, subbed_episode_count + 1):

        if check(episode):

            yield partial(
                lambda anime_id, episode_id: list(
                    iter_streams(session, anime_id, episode_id)
                ),
                anime_id=match.group("anime_id"),
                episode_id=episode,
            ), episode


def metadata_fetcher(session, url, match):
    anime_page = YUGEN + "anime/" + match.group("anime_path")
    anime_page_html = session.get(anime_page).text

    return {
        "titles": [EPISODE_TITLE_REGEX.search(anime_page_html).group(1)],
    }



from warnings import warn

from .base_player import BasePlayer


class AndroidIntentVIEW(BasePlayer):

    intent = "android.intent.action.VIEW"

    warnable_opts = {"headers", "subtitles"}

    def play(self, stream_url, **kwargs):

        args = (
            "am",
            "start",
            "-a",
            self.intent,
            "-d",
            stream_url,
        )

        keys = set(kwargs.keys())

        if self.warnable_opts & keys:
            warn(
                f"Android does not support {', '.join(map(repr, self.warnable_opts & keys))} options. "
                "These options may cause the stream to fail to load or one or more stream attributes to fail."
            )

        self.spawn(args)



import base64
from collections import defaultdict
from functools import lru_cache, partial

import yarl

from ....config import CRUNCHYROLL, SUPERANIME_CRUNCHYROLL
from ... import helpers
from .api import Kamyroll

VRV_RESPONSE_REGEX = helpers.optopt.regexlib.compile(
    r"^#EXT-X-STREAM-INF:.*?RESOLUTION=\d+x(?P<resolution>\d+).*?\n(.+?)$",
    flags=helpers.optopt.regexlib.MULTILINE,
)
CR_GUID_REGEX = helpers.optopt.regexlib.compile(
    r"crunchyroll\.com(?:/.+?)?/series/(.+?)(?:/|$)"
)


REGEX = helpers.optopt.regexlib.compile(
    r"kamyroll://(?:.+?\.)+?crunchyroll\.com(?:/.+?)?/series/(.+?)(?:/|$)"
)


def fix_subtitle_url(url):
    parsed_url = yarl.URL(url)

    token = parsed_url.query.get("token")

    if token is None:
        return url

    url_base = base64.urlsafe_b64decode(token + "=" * (4 - len(token) % 4)).decode()

    return url_base[2:-2]


def fetch_streams(api: Kamyroll, medias: "list"):

    superanime_subtitle = SUPERANIME_CRUNCHYROLL.get("subtitle_language") or ""

    for stream_attrs, media_id in medias:
        streams_data = api.fetch_streams(media_id)

        if not streams_data["streams"]:
            continue

        selected_streams = [
            _
            for _ in streams_data["streams"]
            if _["hardsub_locale"] == superanime_subtitle
        ] or streams_data["streams"]
        subtitles = streams_data["subtitles"]

        for stream in selected_streams:
            yield {
                "stream_url": stream["url"],
                "subtitle": [fix_subtitle_url(_["url"]) for _ in subtitles],
                **stream_attrs,
            }


@lru_cache()
def fetch_episodes(api: Kamyroll, media_id):
    seasons = api.fetch_seasons(media_id)["items"]
    episodes = defaultdict(list)

    for season in seasons:
        for episode in season["episodes"]:
            episode_number = episode["episode_number"] or 0
            episodes[episode_number].append(
                (
                    {
                        "title": f'{episode["title"]} ({season["title"]})',
                    },
                    episode["id"],
                )
            )
    return episodes


def fetcher(session, url, check, match):
    api = Kamyroll(session)
    cr_guid = match.group(1)

    episodes = fetch_episodes(api, cr_guid)

    sorted_episodes = sorted(episodes.items(), key=lambda x: x[0])

    for episode_number, medias in sorted_episodes:
        if check(episode_number):
            yield partial(
                lambda medias: list(fetch_streams(api, medias)),
                medias,
            ), episode_number


def metadata_fetcher(session, url, match):
    api = Kamyroll(session)
    cr_guid = match.group(1)

    metadata = api.fetch_media(cr_guid)

    return {
        "title": [metadata["title"]],
    }



import yarl

from .optopt import regexlib

VRV_RESPONSE_REGEX = regexlib.compile(
    r"^#EXT-X-STREAM-INF:.*?RESOLUTION=\d+x(?P<resolution>\d+).*?\n(.+?)$",
    flags=regexlib.MULTILINE,
)
WIXMP_URL_REGEX = regexlib.compile(
    r"https://.+?/(?P<base>.+?/),(?P<resolutions>(?:\d+p,)+)"
)


def iter_ufph_vrv(session, url: yarl.URL, *, stream_attribs=None):

    if stream_attribs is not None:
        vrv_preferences = stream_attribs.pop("vrv", None)

        if vrv_preferences is not None:
            stream_language = vrv_preferences["stream_type"][9:] or None

            if (
                stream_language
                != vrv_preferences["provider_configuration"]["subtitle_language"]
            ):
                return

    hls_response = session.get(url.human_repr()).text

    for match in VRV_RESPONSE_REGEX.finditer(hls_response):
        stream_url = match.group(2).replace("/index-v1-a1.m3u8", "")

        yield {
            "stream_url": stream_url,
            "quality": int(match.group(1)),
            **(stream_attribs or {}),
        }


def iter_ufph_wixmp(session, url: yarl.URL, *, stream_attribs=None):

    match = WIXMP_URL_REGEX.search(url.human_repr())

    if match is None:
        return

    base_url = match.group("base")

    for resolution in match.group("resolutions").split(",")[:-1]:

        stream_url = f"https://{base_url}{resolution}/mp4/file.mp4"

        yield {
            "stream_url": stream_url,
            "quality": int(resolution[:-1]),
            **(stream_attribs or {}),
        }


UNPACKER_MAPPING = {
    "v.vrv.co": iter_ufph_vrv,
    "repackager.wixmp.com": iter_ufph_wixmp,
}


def iter_unpacked_from_packed_hls(session, url: yarl.URL, *, stream_attribs=None):

    host = url.host

    generator = UNPACKER_MAPPING.get(host, None)

    if generator is None:
        yield {
            "stream_url": url.human_repr(),
            **(stream_attribs or {}),
        }
    else:
        yield from generator(session, url, stream_attribs=stream_attribs)
